from cis_profile.profile import User
from cis_profile.common import WellKnown
from cis_profile.common import DotDict
from cis_profile.common import MozillaDataClassification
from cis_profile.fake_profile import FakeUser

import cis_profile.exceptions

__all__ = [
           User,
           FakeUser,
           DotDict,
           WellKnown,
           MozillaDataClassification,
           DotDict,
           cis_profile.exceptions
]



#!/usr/bin/env python

from cis_profile.common import WellKnown
from cis_profile.common import DotDict
from cis_profile.common import MozillaDataClassification

import cis_crypto.operation
import cis_profile.exceptions
import jose.exceptions
import json
import json.decoder
try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError
import jsonschema
import logging
import os
import time

logger = logging.getLogger(__name__)


class User(object):
    """
    A Mozilla IAM Profile "v2" user structure.
    It is loaded a configuration file (JSON) and dynamically generated.
    If you wish to change the structure, modify the JSON file!

    By default this will load the JSON file with its defaults.

    You can use this like a normal class:
    ```
    from cis_profile import User
    skel_user = User(user_id="bobsmith")
    skel_user.user_id.value = "notbobsmith"
    if skel_user.validate():
        profile = skel_user.as_json()
    ```
    """

    def __init__(self, user_structure_json=None, user_structure_json_file=None,
                 discovery_url='https://auth.mozilla.com/.well-known/mozilla-iam', **kwargs):
        """
        @user_structure_json an existing user structure to load in this class
        @user_structure_json_file an existing user structure to load in this class, from a JSON file
        @discovery_url the well-known Mozilla IAM URL
        @kwargs any user profile attribute name to override on initializing, eg "user_id='test'"
        """
        self.__well_known = WellKnown()

        if (user_structure_json is not None):
            self.load(user_structure_json)
        elif (user_structure_json_file is not None):
            self.load(self.get_profile_from_file(user_structure_json_file))
        else:
            # Load builtin defaults, including default timestamps
            self.load(self.get_profile_from_file('data/user_profile_null.json'))
            self.initialize_timestamps()

        # Insert defaults from kwargs
        for kw in kwargs:
            if kw in self.__dict__.keys():
                try:
                    self.__dict__[kw]['value'] = kwargs[kw]
                except KeyError:
                    self.__dict__[kw]['values'] += [kwargs[kw]]
            else:
                logger.error('Unknown user profile attribute {}'.format(kw))
                raise Exception('Unknown user profile attribute {}'.format(kw))

        self.__signop = cis_crypto.operation.Sign()
        self.__verifyop = cis_crypto.operation.Verify()
        self.__verifyop.well_known = self.__well_known.get_well_known()

    def load(self, profile_json):
        """
        Load an existing JSON profile
        @profile_json: dict (e.g. from json.load() or json.loads())
        """
        logger.debug('Loading profile JSON data structure into class object')
        self.__dict__.update(DotDict(profile_json))

    def get_profile_from_file(self, user_structure_json_path):
        """
        Load the json structure into a 'DotDict' so that attributes appear as addressable object values
        Usually used with load().
        """
        logger.debug('Loading default profile JSON structure from {}'.format(user_structure_json_path))
        if not os.path.isfile(user_structure_json_path):
            dirname = os.path.dirname(os.path.realpath(__file__))
            path = dirname + '/' + user_structure_json_path
        else:
            path = user_structure_json_path
        return DotDict(json.load(open(path)))

    def initialize_timestamps(self):
        # instruct libc that we want UTC
        os.environ['TZ'] = 'UTC'

        now = time.strftime('%Y-%m-%dT%H:%M:%S.000Z')
        logger.debug('Setting all profile metadata fields and profile modification timestamps to now: {}'.format(now))

        for item in self.__dict__:
            if type(self.__dict__[item]) is not DotDict:
                continue
            if 'metadata' in self.__dict__[item]:
                self.__dict__[item]['metadata']['created'] = now
                self.__dict__[item]['metadata']['last_modified'] = now
            else:
                # This is a 2nd level attribute such as `access_information`
                # Note that we do not have a 3rd level so this is sufficient
                for subitem in self.__dict__[item]:
                    if isinstance(self.__dict__[item][subitem], dict) and 'metadata' in self.__dict__[item][subitem]:
                        self.__dict__[item][subitem]['metadata']['created'] = now
                        self.__dict__[item][subitem]['metadata']['last_modified'] = now

        # XXX Hard-coded special profile values
        self.__dict__['last_modified'].value = now
        self.__dict__['created'].value = now

    def update_timestamp(self, req_attr):
        """
        Updates metadata timestamps for that attribute
        @attr a valid user profile attribute
        """
        req_attrs = req_attr.split('.')  # Support subitems/subattributes such as 'access_information.ldap'
        if len(req_attrs) == 1:
            attr = self.__dict__[req_attr]
        else:
            attr = self.__dict__[req_attrs[0]][req_attrs[1]]

        if 'metadata' not in attr:
            raise KeyError("This attribute does not have metadata to update")

        # instruct libc that we want UTC
        os.environ['TZ'] = 'UTC'
        now = time.strftime('%Y-%m-%dT%H:%M:%S.000Z')

        logger.debug('Updating to metadata.last_modified={} for attribute {}'.format(now, req_attr))
        attr['metadata']['last_modified'] = now

    def _clean_dict(self):
        """
        Removes non-user-attrs from internal dict
        """
        user = self.__dict__.copy()
        todel = []
        classname = self.__class__.__name__
        for k in user:
            # Anything that is within the class namespace is whitelisted
            if k.startswith('_{}'.format(classname)):
                todel.append(k)

        for d in todel:
            del user[d]

        return user

    def as_json(self):
        """
        Outputs a JSON version of this user
        Filters out reserved values
        """
        user = self._clean_dict()
        return json.dumps(user)

    def as_dict(self):
        """
        Outputs a real dict version of this user (not a DotDict)
        Filters out reserved values
        """
        user = self._clean_dict()
        return dict(user)

    def filter_scopes(self, scopes=MozillaDataClassification.PUBLIC):
        """
        Filter in place/the current user profile object (self) to only contain attributes with scopes listed in @scopes
        @scopes list of str
        """
        todel = []
        classname = self.__class__.__name__
        for attr in self.__dict__:
            if attr.startswith('_{}'.format(classname)):
                # Don't touch private attrs
                continue
            elif 'metadata' not in self.__dict__[attr]:
                logger.debug('Attribute {} has no metadata, won\'t filter scopes on it'.format(attr))
                continue
            if self.__dict__[attr]['metadata']['classification'] not in scopes:
                todel.append(attr)

        for d in todel:
            logger.debug('Removing attribute {} because it\'s not in {} scopes'.format(attr, scopes))
            del self.__dict__[d]

    def validate(self):
        """
        Validates against a JSON schema
        """

        return jsonschema.validate(self.as_dict(), self.__well_known.get_schema())

    def verify_all_publishers(self, previous_user):
        """
        Verifies all child nodes have an allowed publisher set according to the rules
        @previous_user profile.User object is the previous user we're updating fields from. This allows for checking if
        fields are being updated (value is already set) or created (values are changed from `null`). It defaults to an
        empty profile (with a bunch of `null` values).

        Ex: user.verify_all_publishers(cis_profile.profile.User()) #this checks against a brand new user

        Returns True on success, False if validation fails.
        """
        for item in self.__dict__:
                if type(self.__dict__[item]) is not DotDict:
                    continue
                try:
                    attr = self.__dict__[item]
                    ret = self.verify_can_publish(attr,
                                                  attr_name=item,
                                                  previous_attribute=previous_user.as_dict()[item])
                except (AttributeError, KeyError):
                    # This is the 2nd level attribute match, see also initialize_timestamps()
                    for subitem in self.__dict__[item]:
                        attr = self.__dict__[item][subitem]
                        ret = self.verify_can_publish(attr,
                                                      attr_name=subitem,
                                                      parent_name=item,
                                                      previous_attribute=previous_user.as_dict()[item][subitem])
                if ret is not True:
                    logger.warning('Verification of publisher failed for attribute {}'.format(attr))
                    return False
        return True

    def verify_can_publish(self, attr, attr_name, parent_name=None, previous_attribute=None):
        """
        Verifies that the selected publisher is allowed to change this attribute.
        This works for both first-time publishers ('created' permission) and subsequent updates ('update' permission).

        Note that this does NOT VERIFY SIGNATURE and that you MUST call verify_attribute_signature()
        separately.

        If you do not, any publisher can pass a fake publisher name and this function will answer that the publisher is
        allowed to publish, if the correct one is passed.
        @attr dict requested attribute to verify the publisher of.
        @attr_name str the name of the requested attribute as we cannot look this up.
        @parent_name str the name of the requested attribute's parent name as we cannot look this up.
        @previous_attribute dict the previous attribute if the user is being updated. If None, the value is always
        considered to be "updated" from the current value stored in self.__dict__. Otherwise, it will check against
        the passed value if it's `null` or set, and consider it "created" if it's `null`, "updated" otherwise. Makes
        sense? Good!

        Return bool True on publisher allowed to publish, raise Exception otherwise.
        """
        publisher_name = attr.signature.publisher.name  # The publisher that attempts the change is here
        logger.debug('Verifying that {} is allowed to publish field {}'.format(publisher_name, attr_name))
        operation = 'create'

        # Rules JSON structure:
        # { "create": { "user_id": [ "publisherA", "publisherB"], ...}, "update": { "user_id": "publisherA",... }
        # I know `updators` is not English :)
        rules = self.__well_known.get_publisher_rules()
        if parent_name is None:
            allowed_creators = rules['create'][attr_name]
            allowed_updators = rules['update'][attr_name]
        else:
            try:
                allowed_creators = rules['create'][parent_name][attr_name]
                allowed_updators = rules['update'][parent_name][attr_name]
            except TypeError:  # This is not access_information, this is identities or staff_information
                allowed_creators = rules['create'][parent_name]
                allowed_updators = rules['create'][parent_name]

        # Do we have an attribute to check against?
        if previous_attribute is not None:
            # Creators are only allowed if there is no previous value set
            if (self._attribute_value_set(previous_attribute) is False) and (self._attribute_value_set(attr) is True):
                operation = 'create'
                if publisher_name in allowed_creators:
                    logger.debug('[create] {} is allowed to publish field {}'.format(publisher_name, attr_name))
                    return True
            else:
                operation = 'update'
                # Find if the value changed at all, else don't bother checking and allow it. Otherwise, we'd fail when a
                # publisher tries to validate all fields, but has not actually modified them all.
                # Figure out where values are stored first:
                if 'value' in attr:
                    value = 'value'
                else:
                    value = 'values'

                if attr[value] == previous_attribute[value]:
                    logger.debug('[noop] {} skipped verification for  {} (no changes)'
                                 .format(publisher_name, attr_name))
                    return True

        # No previous attribute set, just check we're allowed to change the field
        else:
            if not self._attribute_value_set(attr):
                operation = 'create'
                logger.debug('[create] {} is allowed to publish field {}'.format(publisher_name, attr_name))
                return True
            elif publisher_name == allowed_updators:
                logger.debug('[update] {} is allowed to publish field {}'.format(publisher_name, attr_name))
                return True

        # None of the checks allowed this change, bail!
        logger.warning('[{}] {} is NOT allowed to publish field {} (value: {}, previous_attribute value: {})'
                       .format(operation, publisher_name, attr_name, attr, previous_attribute))
        raise cis_profile.exceptions.PublisherVerificationFailure('[{}] {} is NOT allowed to publish field {}'
                                                                  .format(operation, publisher_name, attr_name))

    def verify_all_signatures(self):
        """
        Verifies all child nodes with a non-null value's signature against a publisher signature
        """
        for item in self.__dict__:
                if type(self.__dict__[item]) is not DotDict:
                    continue
                try:
                    attr = self.__dict__[item]
                    if self._attribute_value_set(attr):
                        attr = self._verify_attribute_signature(attr)
                except KeyError:
                    # This is the 2nd level attribute match, see also initialize_timestamps()
                    for subitem in self.__dict__[item]:
                        attr = self.__dict__[item][subitem]
                        if self._attribute_value_set(attr):
                            attr = self._verify_attribute_signature(attr)
                if attr is not True:
                    logger.warning('Verification failed for attribute {}'.format(attr))
                    return False
        return True

    def verify_attribute_signature(self, req_attr):
        """
        Verify the signature of an attribute
        @req_attr str this is this user's attribute name, which will be looked up and verified in place
        """
        req_attrs = req_attr.split('.')  # Support subitems/subattributes such as 'access_information.ldap'
        if len(req_attrs) == 1:
            attr = self.__dict__[req_attr]
        else:
            attr = self.__dict__[req_attrs[0]][req_attrs[1]]
        return self._verify_attribute_signature(attr)

    def _verify_attribute_signature(self, attr, publisher_name=None):
        """
        Verify the signature of an attribute
        @attr dict a structure of this user to be verified
        @publisher_name str this is the name of the publisher that should be signing this attribute. If None, the
        publisher_name from the current user structure is used instead and no check is performed.
        """

        if publisher_name is not None and attr['signature']['publisher']['value'] != publisher_name:
            raise cis_profile.exceptions.SignatureVerificationFailure('Incorrect publisher')

        self.__verifyop.load(attr['signature']['publisher']['value'])
        try:
            signed = json.loads(self.__verifyop.jws(publisher_name))
        except jose.exceptions.JWSError as e:
            logger.warning('Attribute signature verification failure: {} ({})'.format(attr, publisher_name))
            raise cis_profile.exceptions.SignatureVerificationFailure('Attribute signature verification failure for {}'
                                                                      '({}) ({})'.format(attr, publisher_name, e))

        # Finally check our object matches the stored data
        attrnosig = attr.copy()
        del attrnosig['signature']

        if signed is None:
            raise cis_profile.exceptions.SignatureVerificationFailure('No data returned by jws() call for '
                                                                      'attribute {}'.format(attr))
        elif signed != attrnosig:
            raise cis_profile.exceptions.SignatureVerificationFailure('Signature data in jws does not match '
                                                                      'attribute data => {} != {}'.format(attrnosig,
                                                                                                          signed))

    def sign_all(self):
        """
        Sign all child nodes with a non-null value(s) OR empty values (strict=False)
        To sign empty values, manually call sign_attribute()
        This requires cis_crypto to be properly setup (i.e. with keys)
        """

        logger.debug('Signing all profile fields that have a value set')
        for item in self.__dict__:
            if type(self.__dict__[item]) is not DotDict:
                continue
            try:
                attr = self.__dict__[item]
                if self._attribute_value_set(attr, strict=False):
                    attr = self._sign_attribute(attr)
            except KeyError:
                # This is the 2nd level attribute match, see also initialize_timestamps()
                for subitem in self.__dict__[item]:
                    attr = self.__dict__[item][subitem]
                    if self._attribute_value_set(attr, strict=False):
                        attr = self._sign_attribute(attr)

    def sign_attribute(self, req_attr):
        """
        Sign a single attribute, including null/empty/unset attributes
        @req_attr str this user's attribute to sign in place
        """
        req_attrs = req_attr.split('.')  # Support subitems/subattributes such as 'access_information.ldap'
        if len(req_attrs) == 1:
            attr = self.__dict__[req_attr]
        else:
            attr = self.__dict__[req_attrs[0]][req_attrs[1]]
        return self._sign_attribute(attr)

    def _attribute_value_set(self, attr, strict=True):
        """
        Checks if an attribute is used/set, ie not null
        @attr dict a complete CIS Profilev2 attribute (such as {'test': {'value': null}})
        @strict bool if True then only null values will be ignored, if False then empty strings/Lists/Dicts will also be
        ignored
        returns: True if the attribute has a value, False if not
        """

        # Note that None is the JSON `null` equivalent (and not "null" is not the string "null")
        if 'value' in attr:
            if attr['value'] is None:
                return False
            elif isinstance(attr['value'], bool):
                return True
            elif not strict and len(attr['value']) == 0:
                return False
        elif 'values' in attr:
            if attr['values'] is None:
                return False
            elif not strict and len(attr['values']) == 0:
                return False
        else:
            raise KeyError(attr)
        return True

    def _sign_attribute(self, attr):
        """
        Perform the actual signature operation
        See also https://github.com/mozilla-iam/cis/blob/profilev2/docs/Profiles.md
        @attr: a CIS Profilev2 attribute
        """
        # Extract the attribute without the signature structure itself
        attrnosig = attr.copy()
        del attrnosig['signature']
        self.__signop.load(attrnosig)

        # Add the signed attribute back to the original complete attribute structure (with the signature struct)
        # This ensure we also don't touch any existing non-publisher signatures
        sigattr = attr['signature']['publisher']
        sigattr['alg'] = 'RS256'  # Currently hardcoded in cis_crypto
        sigattr['typ'] = 'JWS'    # ""
        sigattr['value'] = self.__signop.jws()
        return attr



import json
import os
import requests
import requests.exceptions
import requests_cache
import logging

logger = logging.getLogger(__name__)


class DotDict(dict):
    """
    Convert a dict to a fake class/object with attributes, such as:
    test = dict({"test": {"value": 1}})
    test.test.value = 2
    """
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        try:
            # Python2
            for k, v in self.iteritems():
                self.__setitem__(k, v)
        except AttributeError:
            # Python3
            for k, v in self.items():
                self.__setitem__(k, v)

    def __getattr__(self, k):
        try:
            return dict.__getitem__(self, k)
        except KeyError:
            raise AttributeError("'DotDict' object has no attribute '" + str(k) + "'")

    def __setitem__(self, k, v):
        dict.__setitem__(self, k, DotDict.__convert(v))

    __setattr__ = __setitem__

    def __delattr__(self, k):
        try:
            dict.__delitem__(self, k)
        except KeyError:
            raise AttributeError("'DotDict'  object has no attribute '" + str(k) + "'")

    @staticmethod
    def __convert(o):
        """
        Recursively convert `dict` objects in `dict`, `list`, `set`, and
        `tuple` objects to `DotDict` objects.
        """
        if isinstance(o, dict):
            o = DotDict(o)
        elif isinstance(o, list):
            o = list(DotDict.__convert(v) for v in o)
        elif isinstance(o, set):
            o = set(DotDict.__convert(v) for v in o)
        elif isinstance(o, tuple):
            o = tuple(DotDict.__convert(v) for v in o)
        return o


class MozillaDataClassification(DotDict):
    """
    See https://wiki.mozilla.org/Security/Data_Classification
    Just a simple object-enum - it returns all valid labels per level
    as a list/array.
    """
    UNKNOWN = ['UNKNOWN']
    PUBLIC = ['PUBLIC']
    MOZILLA_CONFIDENTIAL = ['MOZILLA CONFIDENTIAL', 'Mozilla Confidential - Staff and NDA\'d Mozillians Only']
    WORKGROUP_CONFIDENTIAL = ['WORKGROUP CONFIDENTIAL', 'Mozilla Confidential - Specific Work Groups Only']
    INDIVIDUAL_CONFIDENTIAL = ['INDIVIDUAL CONFIDENTIAL', 'Mozilla Confidential - Specific Individuals Only']
    # Well-known Workgroups:
    WELL_KNOWN_WORKGROUPS = ['STAFF_ONLY']
    STAFF_ONLY = ['WORKGROUP CONFIDENTIAL: STAFF ONLY']


class WellKnown(object):
    """
    CIS JSON WellKnown and Schema loader with memory and disk cache and builtin fallback support.
    This object cannot fail to return the schema, but the schema is not garanteed to be up to date in case of network
    issues.

    Tries to get the well-known URL and schema from local cache if the cache has not expired.
    Else, tries to get the schema from well-known URL and cache it.
    Else, uses a library-builtin copy of schema.

    Return: dict Schema dictionary (can be converted to JSON)

    Ex:
    import cis_profile.common.WellKnown
    wk = WellKnown
    print(wk.get_schema())
    <Dict: schema>
    """

    def __init__(self, discovery_url='https://auth.mozilla.com/.well-known/mozilla-iam'):
        self._request_cache = '/var/tmp/cis_request_cache'  # XXX use `get_config` to configure that
        self._request_cache_ttl = 3600
        # Memory cached copies
        self._well_known_json = None
        self._schema_json = None
        self.discovery_url = discovery_url

        logger.debug('Initializing requests_cache TTL={} at {}'.format(self._request_cache_ttl, self._request_cache))
        requests_cache.install_cache(self._request_cache, expire_after=self._request_cache_ttl, backend='sqlite')

    def get_publisher_rules(self):
        """
        Public wrapper for _load_rules
        """
        self._well_known_json = self._load_well_known()
        rules_url = self._well_known_json.get('publishers_rules_uri')
        return self._load_publisher_rules(rules_url)

    def get_schema(self):
        """
        Public wrapper for _load_well_known()
        """
        self._well_known_json = self._load_well_known()
        schema_url = self._well_known_json.get('api').get('data/profile_schema')
        return self._load_schema(schema_url, stype='data/profile.schema')

    def get_core_schema(self):
        """ Deprecated """
        return self.get_schema()

    def get_extended_schema(self):
        """ Deprecated """
        return self.get_schema()

    def get_well_known(self):
        """
        Public wrapper for _load_well_known
        """
        return self._load_well_known()

    def _load_publisher_rules(self, rules_url):
        """
        Get the CIS integration rules
        """
        rules = None
        if rules_url is not None:
            try:
                r = requests.get(rules_url)
                rules = r.json()
                if r.from_cache:
                    logger.debug('Loaded rules data from requests cache')
            except (json.JSONDecodeError, requests.exceptions.ConnectionError) as e:
                logger.debug('Failed to load rules data from rules_url {} ({})'.format(rules_url, e))

        # Fall-back to built-in copy
        if rules is None:
            rules_file = 'data/well-known/mozilla-iam-publisher-rules'
            if not os.path.isfile(rules_file):
                dirname = os.path.dirname(os.path.realpath(__file__))
                path = dirname + '/' + rules_file
            else:
                path = rules_file

            rules = json.load(open(path))
        return rules

    def _load_well_known(self):
        """
        Gets the discovery url's data ("well-known")
        Return dict,None the well-known JSON data copy
        """
        # Memory cache
        if self._well_known_json is not None:
            return self._well_known_json

        try:
            r = requests.get(self.discovery_url)
            self._well_known_json = r.json()
            if r.from_cache:
                logger.debug('Loaded well-known url from requests cache')
        except (json.JSONDecodeError, requests.exceptions.ConnectionError) as e:
            logger.debug('Failed to fetch schema url from discovery {} ({})'.format(self.discovery_url, e))
            logger.debug('Using builtin copy')
            well_known_file = 'data/well-known/mozilla-iam'  # Local fall-back
            if not os.path.isfile(well_known_file):
                dirname = os.path.dirname(os.path.realpath(__file__))
                path = dirname + '/' + well_known_file
            else:
                path = well_known_file
            self._well_known_json = json.load(open(path))

        return self._well_known_json

    def _load_schema(self, schema_url, stype='data/profile.schema'):
        """
        Loads JSON Schema from an URL
        @schema_url: str,None the schema URL
        @stype: str, type of schema to load. This is also the name of the library builtin, local copy.
        Return dict JSON object which is the CIS Profile Schema
        """
        schema = None
        if schema_url is not None:
            try:
                r = requests.get(schema_url)
                schema = r.json()
                if r.from_cache:
                    logger.debug('Loaded schema data from requests cache')
            except (json.JSONDecodeError, requests.exceptions.ConnectionError) as e:
                logger.debug('Failed to load schema from schema_url {} ({})'.format(schema_url, e))

        # That did not work, fall-back to local, built-in copy
        if schema is None:
            # Builtin, hardcoded schema from library
            schema_file = stype
            if not os.path.isfile(schema_file):
                dirname = os.path.dirname(os.path.realpath(__file__))
                path = dirname + '/' + schema_file
            else:
                path = schema_file

            schema = json.load(open(path))
        return schema



#!/usr/bin/env python

import cis_profile.profile
import faker
import faker.providers
import logging
import random

logger = logging.getLogger(__name__)
fake = faker.Faker()


class FakeCISProfileProvider(faker.providers.BaseProvider):
    """
    A provider for cis_profiles fake data
    """

    def user_id(self):
        p = ['ad|Mozilla-LDAP|McDummy',
             'github|23494952',
             'google-oauth2|49459459039398484',
             'oauth2|firefoxaccounts|d0cee943848ef02293dfa',
             'email|5be93248483ee93d']
        return random.choice(p)

    def login_method(self):
        p = ['Mozilla-LDAP',
             'google-oauth2',
             'github',
             'firefoxaccounts',
             'email'
             ]
        return random.choice(p)

    def usernames(self):
        u = {}
        for i in range(0, random.randrange(0, 10)):
            u[fake.words(nb=1)[0]] = fake.user_name()
        return u

    def pub_key(self):
        u = {}
        for i in range(0, random.randrange(0, 3)):
            u[fake.words(nb=1)[0]] = fake.ipv6()
        return u

    def phone(self):
        u = {}
        for i in range(0, random.randrange(0, 3)):
            u[fake.words(nb=1)[0]] = fake.phone_number()
        return u

    def websites(self, s):
        u = {}
        for i in range(0, len(s)):
            u[fake.words(nb=1)[0]] = s[i]
        return u

    def ai(self):
        """
        Fake groups generator
        """
        u = {}
        for i in range(0, random.randrange(0, 10)):
            u[fake.words(nb=1)[0]] = None
        return u

    def worker_type(self):
        p = ['Employee', 'Contractor', 'Intern']
        return random.choice(p)


class FakeUser(cis_profile.profile.User):
    """
    A fake user factory for cis_profile
    @generator int a static seed to always get the same fake profile back
    """

    def __init__(self, generator=None):
        super().__init__()
        if generator is not None:
            fake.seed(generator)

        self.generate(fake)

    def generate(self, fake):
        """
        See also `data/user_profile_null.json`
        """

        fake.add_provider(faker.providers.person)
        fake.add_provider(faker.providers.date_time)
        fake.add_provider(faker.providers.internet)
        fake.add_provider(faker.providers.profile)
        fake.add_provider(faker.providers.lorem)
        fake.add_provider(faker.providers.address)
        fake.add_provider(FakeCISProfileProvider)

        fprofile = fake.profile()

        self.__dict__['user_id']['value'] = fake.user_id()
        self.__dict__['login_method']['value'] = fake.login_method()
        self.__dict__['active']['value'] = fake.boolean()
        self.__dict__['last_modified']['value'] = fake.iso8601()
        self.__dict__['created']['value'] = fake.iso8601()
        self.__dict__['usernames']['values'] = fake.usernames()
        self.__dict__['first_name']['value'] = fake.first_name()
        self.__dict__['last_name']['value'] = fake.last_name()
        self.__dict__['primary_email']['value'] = fprofile['mail']
        if self.login_method == 'Mozilla-LDAP':
            self.__dict__['identities']['mozilla_ldap_id']['value'] = fprofile['mail']
        self.__dict__['identities']['dinopark_id'] = fake.user_name()
        self.__dict__['ssh_public_keys']['values'] = fake.pub_key()
        self.__dict__['pgp_public_keys']['values'] = fake.pub_key()
        if self.login_method == 'Mozilla-LDAP':
            self.__dict__['access_information']['ldap'] = fake.ai()
            self.__dict__['access_information']['hris'] = fake.ai()
        self.__dict__['access_information']['mozilliansorg'] = fake.ai()
        self.__dict__['access_information']['access_provider'] = fake.ai()
        self.__dict__['fun_title']['value'] = fprofile['job']
        self.__dict__['description']['value'] = fake.text(max_nb_chars=200)
        self.__dict__['location']['value'] = fake.country()
        self.__dict__['timezone']['value'] = fake.timezone()
        self.__dict__['languages']['values'] = {'1': 'English'}
        self.__dict__['tags']['values'] = {'1': 'Test'}
        self.__dict__['pronouns']['value'] = None
        self.__dict__['picture'] = fake.image_url()
        self.__dict__['uris']['values'] = fake.websites(fprofile['website'])
        self.__dict__['phone_numbers']['values'] = fake.phone()
        self.__dict__['alternative_name'] = fprofile['name']
        if self.login_method == 'Mozilla-LDAP':
            self.__dict__['staff_information']['manager']['value'] = fake.boolean()
            self.__dict__['staff_information']['director']['value'] = fake.boolean()
            self.__dict__['staff_information']['staff']['value'] = True
            self.__dict__['staff_information']['title']['value'] = fprofile['job']
            self.__dict__['staff_information']['team']['value'] = fake.sentence(nb_words=2)
            self.__dict__['staff_information']['cost_center']['value'] = random.randint(1000, 9000)
            self.__dict__['staff_information']['worker_type']['value'] = fake.worker_type()
            self.__dict__['staff_information']['wpr_desk_number']['value'] = fake.randint(100, 900)
            self.__dict__['staff_information']['office_location']['value'] = fake.city()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = ['jsonschema', 'requests', 'requests-cache', 'graphene', 'Faker']
test_requirements = ['pytest', 'pytest-watch', 'pytest-cov', 'flake8', 'flask', 'flask_graphql',
                     'flask_restful']
setup_requirements = [
    'pytest-runner',
    'setuptools>=40.5.0'
]

extras = {'test': test_requirements}

setup(
    name="cis_profile",
    version="0.0.3.1",
    author="Guillaume Destuynder",
    author_email="kang@mozilla.com",
    description="Mozilla IAM user profile ('v2') class utility.",
    long_description=long_description,
    url="https://github.com/mozilla-iam/cis",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    license="Mozilla Public License 2.0",
    include_package_data=True,
    packages=find_packages(include=['cis_profile', 'cis_profile/data', 'cis_profile/data/well-known']),
    package_data={
        'cis_profile': [
            '*.json',
            'data/*.schema',
            'data/*.json',
            'data/well-known/mozilla-iam',
            'data/well-known/mozilla-iam-publisher-rules'
        ]
    },
    setup_requires=setup_requirements,
    tests_require=test_requirements,
    extras_require=extras,
    test_suite='tests',
    zip_safe=True
)



"""
Graphene objects that represent the User Profile Schema in Graphene-compatible format.
These can be loaded by libraries or code understanding `graphene.ObjectType`
Object have to be loaded into a `graphene.Schema(graphene.Query(OurTopLevelObject))` style call
"""
import datetime
import graphene

from graphene.types import Scalar
from graphql.language import ast


# XXX use a schema generator for all things schema instead of this as other classes use the schema file as input
# while this part of the code hardcodes the schema. Generating classes or classes seems to be a little hard to
# read or understand for mere mortals, eg:
# class list: self.classes.append(type("classname", (object), {'val': 'test', 'func': func})
# This causes issues particularly if you need to draw a tree of interdependent classes (ie `func` above is part of a
# class that is not-yet-generated, thus a tree of dependencies is first resolved, then class lists generated, which
# seems frankly overkill
# See commit c38f00975896cb60a1467e3315d2291dd5f31657 for a Schema generator head-start

# Required non-scalar to scalar transforms
class DateTime(Scalar):
    """
    Graphene Scaler for our date time format
    """
    @staticmethod
    def serialize(dt):
        return dt.isoformat()

    @staticmethod
    def parse_literal(node):
        if isinstance(node, ast.StringValue):
            return datetime.datetime.strptime(
                node.value, "%Y-%m-%dT%H:%M:%S.%f")

    @staticmethod
    def parse_value(value):
        return datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f")


# Profile enums
class Alg(graphene.Enum):
    """
    Supported signature algorithms
    """
    HS256 = 'HS256'
    RS256 = 'RS256'
    RSA = 'RSA'
    ED25519 = 'ED25519'


class Typ(graphene.Enum):
    """
    Supported signature types
    """
    JWT = 'JWT'
    PGP = 'PGP'


class Classification(graphene.Enum):
    """
    Mozilla data classification
    """
    MOZILLA_CONFIDENTIAL = 'MOZILLA CONFIDENTIAL'
    PUBLIC = 'PUBLIC'
    WORKGROUP_CONFIDENTIAL = 'WORKGROUP CONFIDENTIAL'
    STAFF_ONLY = 'WORKGROUP CONFIDENTIAL: STAFF ONLY'
    INDIVIDUAL_CONFIDENTIAL = 'INDIVIDUAL CONFIDENTIAL'


# Standard properties
class Display(graphene.Enum):
    """
    DinoPark visibility/display intent
    """
    public = 'public'
    authenticated = 'authenticated'
    vouched = 'vouched'
    ndaed = 'ndaed'
    staff = 'staff'
    private = 'private'
    null = 'null'


class Metadata(graphene.ObjectType):
    classification = graphene.Field(Classification)
    last_modified = DateTime()
    created = DateTime()
    verified = graphene.Boolean()
    display = graphene.Field(Display)


class SignatureField(graphene.ObjectType):
    alg = graphene.Field(Alg)
    typ = graphene.Field(Typ)
    name = graphene.String()
    value = graphene.String()


class Signature(graphene.ObjectType):
    publisher = graphene.Field(SignatureField)
    additional = graphene.List(SignatureField)


class StandardProperty(graphene.ObjectType):
    signature = graphene.Field(Signature)
    metadata = graphene.Field(Metadata)


class StandardAttributeString(StandardProperty):
    value = graphene.String()


class StandardAttributeList(StandardProperty):
    values = graphene.List(graphene.String)


class StandardAttributeBoolean(StandardProperty):
    value = graphene.Boolean()


class StandardAttributeDateTime(StandardProperty):
    value = DateTime()


class StandardAttributeFieldList(StandardProperty):
    """
    graphene requires Fields (dicts) to have well-known keys
    work-around this by providing a list of fields through a graphene List
    """
    values = graphene.List(graphene.String)

    def resolve_values(self, info, *kwargs):
        values = self.get('values')
        if values:
            return values.items()
        else:
            return None


# Profile advanced properties
class AccessInformation(StandardProperty):
    ldap = graphene.Field(StandardAttributeFieldList)
    mozilliansorg = graphene.Field(StandardAttributeFieldList)
    hris = graphene.Field(StandardAttributeFieldList)
    access_provider = graphene.Field(StandardAttributeFieldList)


# Profiles
class StaffInformation(StandardProperty):
    manager = graphene.Field(StandardAttributeString)
    director = graphene.Field(StandardAttributeString)
    staff = graphene.Field(StandardAttributeString)
    title = graphene.Field(StandardAttributeString)
    team = graphene.Field(StandardAttributeString)
    cost_center = graphene.Field(StandardAttributeString)
    worker_type = graphene.Field(StandardAttributeString)
    wpr_desk_number = graphene.Field(StandardAttributeString)
    office_location = graphene.Field(StandardAttributeString)


class Identities(StandardProperty):
    github_id_v3 = graphene.Field(StandardAttributeString)
    github_id_v4 = graphene.Field(StandardAttributeString)
    dinopark_id = graphene.Field(StandardAttributeString)
    mozilliansorg_id = graphene.Field(StandardAttributeString)
    bugzilla_mozilla_org_id = graphene.Field(StandardAttributeString)
    mozilla_ldap_id = graphene.Field(StandardAttributeString)
    mozilla_posix_id = graphene.Field(StandardAttributeString)
    google_oauth2_id = graphene.Field(StandardAttributeString)
    firefox_accounts_id = graphene.Field(StandardAttributeString)


## Profiles
class Profile(graphene.ObjectType):
    """
    IAM Profile v2
    """
    user_id = graphene.Field(StandardAttributeString)
    login_method = graphene.Field(StandardAttributeString)
    active = graphene.Field(StandardAttributeBoolean)
    last_modified = graphene.Field(StandardAttributeDateTime)
    created = graphene.Field(StandardAttributeDateTime)
    usernames = graphene.Field(StandardAttributeList)
    first_name = graphene.Field(StandardAttributeString)
    last_name = graphene.Field(StandardAttributeString)
    primary_email = graphene.Field(StandardAttributeString)
    identities = graphene.Field(Identities)
    ssh_public_keys = graphene.Field(StandardAttributeFieldList)
    pgp_public_keys = graphene.Field(StandardAttributeFieldList)
    access_information = graphene.Field(AccessInformation)
    fun_title = graphene.Field(StandardAttributeString)
    description = graphene.Field(StandardAttributeString)
    location = graphene.Field(StandardAttributeString)
    timezone = graphene.Field(StandardAttributeString)
    language = graphene.Field(StandardAttributeString)
    tags = graphene.Field(StandardAttributeList)
    pronouns = graphene.Field(StandardAttributeString)
    picture = graphene.Field(StandardAttributeString)
    uris = graphene.Field(StandardAttributeFieldList)
    phone_numbers = graphene.Field(StandardAttributeFieldList)
    alternative_name = graphene.Field(StandardAttributeString)
    staff_information = graphene.Field(StaffInformation)
